from spotdl.download.tracking_file_handler import DownloadTracker
from spotdl.download.progress_ui_handler import DisplayManager
from spotdl.download.ffmpeg import convert, has_correct_version
from spotdl.download.embed_metadata import set_id3_data
from spotdl.download.downloader import DownloadManager



from spotdl.providers.metadata_provider import from_url
from spotdl.providers import ytm_provider as youtube_music
from spotdl.providers import lyrics_providers
from spotdl.providers.provider_utils import (
    _create_song_title,
    _match_percentage,
    _parse_duration,
)



import asyncio
from pathlib import Path
from types import SimpleNamespace

import pytest

from spotdl.download.downloader import DownloadManager
from spotdl.search import SongObject
from spotdl.download import ffmpeg, downloader


def create_song_obj(name: str = None, artists_input: list = None) -> SongObject:
    if name == None:
        song_name = "test song"
    else:
        song_name = name

    if artists_input == None:
        artist_objs = list(map(lambda x: {"name": x}, ["test artist"]))
    else:
        artist_objs = list(map(lambda x: {"name": x}, artists_input))

    raw_track_meta = {
        "name": song_name,
        "album": {
            "name": "test album",
            "artists": artist_objs,
            "release_date": "2021",
            "images": [
                {"url": "https://i.ytimg.com/vi_webp/iqKdEhx-dD4/hqdefault.webp"}
            ],
        },
        "artists": artist_objs,
        "track_number": "1",
        "genres": ["test genre"],
    }

    raw_album_meta = {"genres": ["test genre"]}
    raw_artist_meta = {"genres": ["test artist genre"]}

    return SongObject(
        raw_track_meta,
        raw_album_meta,
        raw_artist_meta,
        "https://www.youtube.com/watch?v=Th_C95UMegc",
        "test lyrics",
        None,
    )


class FakeProcess:
    """Instead of running ffmpeg, just fake it"""

    def __init__(self, *args):
        command = list(*args)
        self._input = Path(command[command.index("-i") + 1])
        self._output = Path(command[-1])

    async def communicate(self):
        """
        Ensure that the file has been download, and create empty output file,
        to avoid infinite loop.
        """
        assert self._input.is_file()
        self._output.open("w").close()
        return (None, None)

    async def wait(self):
        return None

    @property
    def returncode(self):
        return 0


async def fake_create_subprocess_exec(*args, stdout=None, stderr=None):
    return FakeProcess(args)


@pytest.fixture()
def setup(tmpdir, monkeypatch):
    monkeypatch.chdir(tmpdir)
    monkeypatch.setattr(ffmpeg, "has_correct_version", lambda *_: True)
    monkeypatch.setattr(
        asyncio.subprocess, "create_subprocess_exec", fake_create_subprocess_exec
    )
    monkeypatch.setattr(downloader, "set_id3_data", lambda *_: None)
    data = SimpleNamespace()
    data.directory = tmpdir
    yield data


@pytest.mark.vcr()
def test_download_single_song(setup):
    song_obj = create_song_obj()
    with DownloadManager() as dm:
        dm.download_single_song(song_obj)

    assert [file.basename for file in setup.directory.listdir() if file.isfile()] == [
        "test artist - test song.mp3"
    ]


@pytest.mark.vcr()
def test_download_long_artists_song(setup):
    # ! Generates a long list of artists, numbered 1 to 260, to trigger filename length cases
    artists = [str(i) for i in range(260)]
    song_obj = create_song_obj(artists_input=artists)
    with DownloadManager() as dm:
        dm.download_single_song(song_obj)

    assert [file.basename for file in setup.directory.listdir() if file.isfile()] == [
        "0 - test song.mp3"
    ]


@pytest.mark.vcr()
def test_download_long_name_song(setup):
    # ! Generates a long title name,numbered 1 to 260, to trigger filename length cases
    # ! In this case the program cannot save the song, and fails with an OSError
    song = "a" * 260
    song_obj = create_song_obj(name=song)

    with pytest.raises(OSError):
        with DownloadManager() as dm:
            dm.download_single_song(song_obj)


def test_download_multiple_songs(pytestconfig, setup):
    if not "--disable-vcr" in pytestconfig.invocation_params.args:
        # this test is very similar to the other one, and the http request
        # seems not deterministic so it can't be reliably capture into cassette,
        # therefore run this test only when VCR is disabled
        pytest.skip()

    song_objs = [
        create_song_obj(name="song1"),
        create_song_obj(name="song2"),
        create_song_obj(name="song3"),
    ]
    with DownloadManager() as dm:
        dm.download_multiple_songs(song_objs)

    assert sorted(
        [file.basename for file in setup.directory.listdir() if file.isfile()]
    ) == sorted(
        [
            "test artist - song1.mp3",
            "test artist - song2.mp3",
            "test artist - song3.mp3",
        ]
    )



# ===============
# === Imports ===
# ===============

from pytube.__main__ import YouTube
from unidecode import unidecode
from pytube import Search

# ! Just for static typing
from typing import List, Optional

from spotdl.providers.provider_utils import (
    _match_percentage,
    _create_song_title,
)


def search_and_get_best_match(
    song_name: str,
    song_artists: List[str],
    song_duration: int,
    isrc: str,
) -> Optional[str]:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    `str` `song_album_name` : name of song's album

    `int` `song_duration` : duration of the song

    `str` `isrc` :  code for identifying sound recordings and music video recordings

    RETURNS `str` : link of the best match
    """

    # if isrc is not None then we try to find song with it
    if isrc is not None:
        isrc_results = Search(isrc).results

        if isrc_results and len(isrc_results) == 1:
            isrc_result = isrc_results[0]

            if isrc_result is not None and isrc_result.watch_url is not None:
                return isrc_result.watch_url

    song_title = _create_song_title(song_name, song_artists).lower()

    # Query YTM by songs only first, this way if we get correct result on the first try
    # we don't have to make another request to ytmusic api that could result in us
    # getting rate limited sooner
    results = Search(song_title).results

    if results is None:
        print("Couldn't find the song on YouTube")
        return None

    # Order results
    results = _order_yt_results(results, song_name, song_artists, song_duration)

    # No matches found
    if len(results) == 0:
        return None

    result_items = list(results.items())

    # Sort results by highest score
    sorted_results = sorted(result_items, key=lambda x: x[1], reverse=True)

    # ! In theory, the first 'TUPLE' in sorted_results should have the highest match
    # ! value, we send back only the link
    return sorted_results[0][0]


def _order_yt_results(
    results: List[YouTube],
    song_name: str,
    song_artists: List[str],
    song_duration: int,
) -> dict:

    # Assign an overall avg match value to each result
    links_with_match_value = {}

    for result in results:
        # ! skip results without videoId, this happens if you are country restricted or
        # ! video is unavailabe
        if result.video_id is None:
            continue

        lower_song_name = song_name.lower()
        lower_result_name = result.title.lower()

        sentence_words = lower_song_name.replace("-", " ").split(" ")

        common_word = False

        # ! check for common word
        for word in sentence_words:
            if word != "" and word in lower_result_name:
                common_word = True

        # ! if there are no common words, skip result
        if common_word is False:
            continue

        # Find artist match
        # ! match  = (no of artist names in result) / (no. of artist names on spotify) * 100
        artist_match_number = 0

        # ! we use fuzzy matching because YouTube spellings might be mucked up
        # ! i.e if video
        for artist in song_artists:
            # ! something like _match_percentage('rionos', 'aiobahn, rionos Motivation
            # ! (remix)' would return 100, so we're absolutely corrent in matching
            # ! artists to song name.
            if _match_percentage(
                unidecode(artist.lower()), unidecode(result.title).lower(), 85
            ):
                artist_match_number += 1

        # ! Skip if there are no artists in common, (else, results like 'Griffith Swank -
        # ! Madness' will be the top match for 'Ruelle - Madness')
        if artist_match_number == 0:
            continue

        artist_match = (artist_match_number / len(song_artists)) * 100
        song_title = _create_song_title(song_name, song_artists).lower()
        name_match = round(
            _match_percentage(
                unidecode(result.title.lower()), unidecode(song_title), 60
            ),
            ndigits=3,
        )

        # skip results with name match of 0, these are obviously wrong
        # but can be identified as correct later on due to other factors
        # such as time_match or artist_match
        if name_match == 0:
            continue

        # Find duration match
        # ! time match = 100 - (delta(duration)**2 / original duration * 100)
        # ! difference in song duration (delta) is usually of the magnitude of a few
        # ! seconds, we need to amplify the delta if it is to have any meaningful impact
        # ! wen we calculate the avg match value
        delta = result.length - song_duration  # ! check this
        non_match_value = (delta ** 2) / song_duration * 100

        time_match = 100 - non_match_value

        average_match = (artist_match + name_match + time_match) / 3

        # the results along with the avg Match
        links_with_match_value[result.watch_url] = average_match

    return links_with_match_value



from pathlib import Path
from typing import Optional, List

from spotdl.search import SongObject, song_gatherer


class DownloadTracker:
    def __init__(self):
        self.song_list = []
        self.save_file: Optional[Path] = None

    def load_tracking_file(self, tracking_file_path: str) -> None:
        """
        `str` `tracking_file_path` : path to a .spotdlTrackingFile

        RETURNS `~`

        reads songsObj's from disk and prepares to track their download
        """

        # Attempt to read .spotdlTrackingFile, raise exception if file can't be read
        tracking_file = Path(tracking_file_path).resolve()
        if not tracking_file.is_file():
            raise FileNotFoundError(
                f"no such tracking file found: {tracking_file_path}"
            )

        with tracking_file.open("rb") as file_handle:
            song_data_dumps = eval(file_handle.read().decode())

        # Save path to .spotdlTrackingFile
        self.save_file = tracking_file

        # convert song data dumps to songObj's
        # ! see, songGatherer.get_data_dump and songGatherer.from_dump for more details
        for dump in song_data_dumps:
            self.song_list.append(song_gatherer.from_dump(dump))

    def load_song_list(self, song_list: List[SongObject]) -> None:
        """
        `list<songOjb>` `song_list` : songObj's being downloaded

        RETURNS `~`

        prepares to track download of provided songObj's
        """

        self.song_list = song_list

        self.backup_to_disk()

    def get_song_list(self) -> List[SongObject]:
        """
        RETURNS `list<songObj>

        get songObj's representing songs yet to be downloaded
        """
        return self.song_list

    def backup_to_disk(self):
        """
        RETURNS `~`

        backs up details of songObj's yet to be downloaded to a .spotdlTrackingFile
        """
        # remove tracking file if no songs left in queue
        # ! we use 'return None' as a convenient exit point
        if len(self.song_list) == 0:
            if self.save_file and self.save_file.is_file():
                self.save_file.unlink()

            return None

        # prepare datadumps of all songObj's yet to be downloaded
        song_data_dumps = [song.data_dump for song in self.song_list]

        # ! the default naming of a tracking file is $nameOfFirstSOng.spotdlTrackingFile,
        # ! it needs a little fixing because of disallowed characters in file naming
        if not self.save_file:
            song_name = self.song_list[0].song_name

            song_name = "".join(char for char in song_name if char not in "/?\\*|<>")

            song_name = song_name.replace('"', "'").replace(":", " - ")

            self.save_file = Path(song_name + ".spotdlTrackingFile")

        # save encoded string to a file
        with open(self.save_file, "wb") as file_handle:
            file_handle.write(str(song_data_dumps).encode())

    def notify_download_completion(self, song_object: SongObject) -> None:
        """
        `songObj` `songObj` : songObj representing song that has been downloaded

        RETURNS `~`

        removes given songObj from download queue and updates .spotdlTrackingFile
        """

        if song_object in self.song_list:
            self.song_list.remove(song_object)

        self.backup_to_disk()

    def clear(self):
        self.song_list = []
        self.save_file = None



from requests import get
from bs4 import BeautifulSoup
from typing import List
from urllib.parse import quote

user_agent = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 \
(KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"
}


def get_lyrics_genius(song_name: str, song_artists: List[str]) -> str:
    """
    `str` `song_name` : name of song
    `list<str>` `song_artists` : list containing name of contributing artists
    RETURNS `str`: Lyrics of the song.
    Gets the lyrics of the song.
    """
    headers = {
        "Authorization": "Bearer alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG",
    }
    headers.update(user_agent)

    api_search_url = "https://api.genius.com/search"
    artist_str = ", ".join(
        artist for artist in song_artists if artist.lower() not in song_name.lower()
    )
    search_query = f"{song_name} {artist_str}"

    api_response = get(api_search_url, params={"q": search_query}, headers=headers)
    if not api_response.ok:
        return ""
    api_json = api_response.json()

    try:
        song_id = api_json["response"]["hits"][0]["result"]["id"]
    except (IndexError, KeyError):
        return ""

    song_api_url = f"https://api.genius.com/songs/{song_id}"
    api_response = get(song_api_url, headers=headers)
    if not api_response.ok:
        return ""
    api_json = api_response.json()

    song_url = api_json["response"]["song"]["url"]
    genius_page = get(song_url, headers=user_agent)
    if not genius_page.ok:
        return ""

    soup = BeautifulSoup(genius_page.text.replace("<br/>", "\n"), "html.parser")
    lyrics_div = soup.select_one("div.lyrics")

    if lyrics_div is not None:
        return lyrics_div.get_text().strip()

    lyrics_containers = soup.select("div[class^=Lyrics__Container]")
    lyrics = "\n".join(con.get_text() for con in lyrics_containers)
    return lyrics.strip()


def get_lyrics_musixmatch(
    song_name: str, song_artists: List[str], track_search=False
) -> str:
    """
    `str` `song_name` : Name of song
    `list<str>` `song_artists` : list containing name of contributing artists
    `bool` `track_search` : if `True`, search the musixmatch tracks page.
    RETURNS `str`: Lyrics of the song.
    Gets the lyrics of the song.
    """
    # remove artist names that are already in the song_name
    # we do not use SongObject.create_file_name beacause it
    # removes '/' etc from the artist and song names.
    artists_str = ", ".join(
        artist for artist in song_artists if artist.lower() not in song_name.lower()
    )

    # quote the query so that it's safe to use in a url
    # e.g "Au/Ra" -> "Au%2FRa"
    query = quote(f"{song_name} - {artists_str}", safe="")

    # search the `tracks page` if track_search is True
    if track_search:
        query += "/tracks"

    search_url = f"https://www.musixmatch.com/search/{query}"
    search_resp = get(search_url, headers=user_agent)
    if not search_resp.ok:
        return ""

    search_soup = BeautifulSoup(search_resp.text, "html.parser")
    song_url_tag = search_soup.select_one("a[href^='/lyrics/']")

    # song_url_tag being None means no results were found on the
    # All Results page, therefore, we use `track_search` to
    # search the tracks page.
    if song_url_tag is None:
        # track_serach being True means we are already searching the tracks page.
        if track_search:
            return ""

        lyrics = get_lyrics_musixmatch(song_name, song_artists, track_search=True)
        return lyrics

    song_url = "https://www.musixmatch.com" + str(song_url_tag.get("href", ""))
    lyrics_resp = get(song_url, headers=user_agent)
    if not lyrics_resp.ok:
        return ""

    lyrics_soup = BeautifulSoup(lyrics_resp.text, "html.parser")
    lyrics_paragraphs = lyrics_soup.select("p.mxm-lyrics__content")
    lyrics = "\n".join(i.get_text() for i in lyrics_paragraphs)

    return lyrics



import re

from pathlib import Path
from typing import List

from rapidfuzz import fuzz


def _match_percentage(str1: str, str2: str, score_cutoff: float = 0) -> float:
    """
    A wrapper around `rapidfuzz.fuzz.partial_ratio` to handle UTF-8 encoded
    emojis that usually cause errors

    `str` `str1` : a random sentence
    `str` `str2` : another random sentence
    `float` `score_cutoff` : minimum score required to consider it a match
        returns 0 when similarity < score_cutoff

    RETURNS `float`
    """

    # ! this will throw an error if either string contains a UTF-8 encoded emoji
    try:
        return fuzz.partial_ratio(str1, str2, score_cutoff=score_cutoff)

    # ! we build new strings that contain only alphanumerical characters and spaces
    # ! and return the partial_ratio of that
    except:  # noqa:E722
        new_str1 = "".join(
            each_letter
            for each_letter in str1
            if each_letter.isalnum() or each_letter.isspace()
        )

        new_str2 = "".join(
            each_letter
            for each_letter in str2
            if each_letter.isalnum() or each_letter.isspace()
        )

        return fuzz.partial_ratio(new_str1, new_str2, score_cutoff=score_cutoff)


def _parse_duration(duration: str) -> float:
    """
    Convert string value of time (duration: "25:36:59") to a float value of seconds (92219.0)
    """
    try:
        # {(1, "s"), (60, "m"), (3600, "h")}
        mapped_increments = zip([1, 60, 3600], reversed(duration.split(":")))
        seconds = sum(multiplier * int(time) for multiplier, time in mapped_increments)
        return float(seconds)

    # ! This usually occurs when the wrong string is mistaken for the duration
    except (ValueError, TypeError, AttributeError):
        return 0.0


def _create_song_title(song_name: str, song_artists: List[str]) -> str:
    joined_artists = ", ".join(song_artists)
    return f"{joined_artists} - {song_name}"


def _sanitize_filename(input_str: str) -> str:
    output = input_str

    # ! this is windows specific (disallowed chars)
    output = "".join(char for char in output if char not in "/?\\*|<>")

    # ! double quotes (") and semi-colons (:) are also disallowed characters but we would
    # ! like to retain their equivalents, so they aren't removed in the prior loop
    output = output.replace('"', "'").replace(":", "-")

    return output


def _get_smaller_file_path(input_song, output_format: str) -> Path:
    # Only use the first artist if the song path turns out to be too long
    smaller_name = f"{input_song.contributing_artists[0]} - {input_song.song_name}"

    smaller_name = _sanitize_filename(smaller_name)

    try:
        return Path(f"{smaller_name}.{output_format}").resolve()
    except OSError:
        # Expected to happen in the rare case when the saved path is too long,
        # even with the short filename
        raise OSError("Cannot save song due to path issues.")


def _get_converted_file_path(song_obj, output_format: str = None) -> Path:

    # ! we eliminate contributing artist names that are also in the song name, else we
    # ! would end up with things like 'Jetta, Mastubs - I'd love to change the world
    # ! (Mastubs REMIX).mp3' which is kinda an odd file name.

    # also make sure that main artist is included in artistStr even if they
    # are in the song name, for example
    # Lil Baby - Never Recover (Lil Baby & Gunna, Drake).mp3

    artists_filtered = []

    if output_format is None:
        output_format = "mp3"

    for artist in song_obj.contributing_artists:
        if artist.lower() not in song_obj.song_name:
            artists_filtered.append(artist)
        elif artist.lower() is song_obj.contributing_artists[0].lower():
            artists_filtered.append(artist)

    artist_str = ", ".join(artists_filtered)

    converted_file_name = _sanitize_filename(
        f"{artist_str} - {song_obj.song_name}.{output_format}"
    )

    converted_file_path = Path(converted_file_name)

    # ! Checks if a file name is too long (256 max on both linux and windows)
    try:
        if len(str(converted_file_path.resolve().name)) > 256:
            print("Path was too long. Using Small Path.")
            return _get_smaller_file_path(song_obj, output_format)
    except OSError:
        return _get_smaller_file_path(song_obj, output_format)

    return converted_file_path


def _parse_path_template(path_template, song_object, output_format, short=False):
    converted_file_name = path_template

    converted_file_name = converted_file_name.format(
        artist=_sanitize_filename(song_object.contributing_artists[0]),
        title=_sanitize_filename(song_object.song_name),
        album=_sanitize_filename(song_object.album_name),
        playlist=_sanitize_filename(song_object.playlist_name) if song_object.playlist_name else "",
        artists=_sanitize_filename(
            ", ".join(song_object.contributing_artists)
            if short is False
            else song_object.contributing_artists[0]
        ),
        ext=_sanitize_filename(output_format),
    )

    if len(converted_file_name) > 250:
        return _parse_path_template(
            path_template, song_object, output_format, short=True
        )

    converted_file_path = Path(converted_file_name)

    return converted_file_path
